package biggests.tables;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.table.*;

import biggests.gui.*;
import biggests.utils.*;
import de.schlichtherle.io.*;
import smadeira.biclustering.*;
import smadeira.ontologizer.*;

/**
 * <p>Title: Colors Table</p>
 *
 * <p>Description:
 *                 <p>Defines a table of colors with the following data:<br/>
 *                 - header -> conditions' names (more widely, <code>String</code>s)<br/>
 *                 - first column -> genes' names (more widely, <code>String</code>s)<br/>
 *                 - all other columns' cells -> gene expression values
 *                 (more widely, <code>float</code> values)</p>
 *
 *                 <p>A table of colors has a vertical text header and each
 *                 of its cells is colored according to the float value that
 *                 holds in it.</p>
 *
 * <p>Each one of the matrix's cells is colored according to its
 * expression value.</p>
 * <p>
 * The cells whose value is situated in the interval:<br/>
 * <b>[minValue, <code>meanValue</code>[</b> are colored green
 * (different shades of green, according to the exact value)<br/>
 * <b>{<code>meanValue</code>}</b> are colored black<br/>
 * <b>]<code>meanValue</code>, maxValue]</b> are colored red
 * (different shades of red, according to the exact value)
 * </p>
 * <p>
 * This table also displays the terms annotated with each gene
 * (when a specific table's row is clicked by the user).
 * </p>
 *
 *                 <p>This kind of table implements sorting functionality
 *                 on the columns, i.e., allows the reordering of the rows,
 *                 sorting them by a specific column's cells values in
 *                 ascending or descending order.</p>
 * </p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonalves
 */
/*
 * A colored table looks like this:<br/>
 * ---------------------------------------------------------------------<br/>
 * | Genes Names | C1 | C2 | C3 | .....................................|<br/>
 * |-------------------------------------------------------------------|<br/>
 * | Gene1       | o  | o  | o  | .................................... |<br/>
 * | Gene2       | o  | o  | o  | .................................... |<br/>
 * | Gene3       | o  | o  | o  | .................................... |<br/>
 * | Gene4       | o  | o  | o  | .................................... |br/>
 * ---------------------------------------------------------------------<br/>
 * where 'o' represents a colored cell. Conditions' names are written with
 * vertical orientation.
 */
public class TableColors extends TableValues {
    /**
     * The mean of the node object matrix's values. It is used to calculate the
     * red or green component value of each colored matrix's cell.<br/>
     * The mean value is calculated like this:<br/>
     * <b>(maxValue + minValue)/2</b><br/>
     * in which the maxValue is the maximum value of all the node object matrix
     * cells' values and the minValue is the minimum value of all.
     */
    protected float meanValue;

    /**
     * Used to calculate the red or green component value of each colored
     * matrix's cell.<br/>
     * The divider value is calculated like this:<br/>
     * <b>(maxValue - meanValue)/255</b><br/>
     * in which maxValue is the maximum value of all the node object matrix
     * cells' values, the minValue is the minimum value of all and meanValue
     * is the mean value (the value which is exactly in the middle between
     * maximum and minimum values).
     */
    protected float divider;

    /**
     * Creates a new colored table with data from the the given node object's
     * matrix.
     *
     * @param nodeObject the <code>NodeObjectInterface</code> from which the
     *                   colored table will be created
     * @param go the <code>GOFrontEnd</code>
     * @param parent the GUI's parent <code>Component</code>
     * @throws InvalidNodeObjectTypeException when the type of the node
     *                   object is not valid
     *
     * @see #TableColors(NodeObjectInterface nodeObject)
     */
    public TableColors(NodeObjectInterface nodeObject, GOFrontEnd go,
                       Component parent) throws InvalidNodeObjectTypeException {
        this(nodeObject);
        if (nodeObject instanceof IMatrix) {
            goProperties = ((IMatrix) nodeObject).getGeneOntologyProperties();
        } else
        if (nodeObject instanceof Biclustering) {
            goProperties = ((IMatrix) ((Biclustering) nodeObject).getMatrix()).
                           getGeneOntologyProperties();
        } else {
            // nodeObject instanceof Bicluster
            if (go != null)
                goProperties = ((Bicluster) nodeObject).
                               computeGeneOntologyTerms(go, false);
        }
        this.goFrontEnd = go;
        this.parent = parent;
    }

    /**
     * Creates a new colored table with data from the the given node object's
     * matrix.
     *
     * @param nodeObject the <code>NodeObjectInterface</code> from which the
     *                   colored table will be created
     * @throws InvalidNodeObjectTypeException when the type of the node
     *                   object is not valid
     *
     * @see #TableColors(Object[][] rowData, Object[] columnNames)
     * @see TableValues#constructRowData(NodeObjectInterface nodeObject)
     * @see TableValues#constructColumnNames(NodeObjectInterface nodeObject)
     */
    public TableColors(NodeObjectInterface nodeObject) throws
            InvalidNodeObjectTypeException {
        this(constructRowData(nodeObject), constructColumnNames(nodeObject), true);
    }


    /**
     * Creates a new colored table with the given row data and columns' names.
     *
     * @param rowData the <code>Object[][]</code> array with the data to fill
     *                the rows of the table
     * @param columnNames the <code>Object[][]</code> array with the data to
     *                    fill the header of the table
     * @param meanDivider <code>boolean</code> <code>true</code> if mean and
     *                    divider values should be calculated over the
     *                    <code>rowData</code>
     * @throws InvalidNodeObjectTypeException when the type of the node
     *                   object is not valid
     * @see #calcMeanDividerValues(Object[][] rowData)
     * @see TableValues#TableValues(TableSorter tableSorterModel)
     * @see javax.swing.table.DefaultTableModel#DefaultTableModel(Object[][] rowData, Object[] columnNames)
     */
    protected TableColors(Object[][] rowData, Object[] columnNames,
                          boolean meanDivider) {
        super(new TableSorterVerticalHeader(new DefaultTableModel(rowData,
                columnNames)));

        int size;
        if ((rowData[0].length - 1) > 45) {
            size = 12;
            this.setRowHeight(size);
        } else {
            size = 20;
        }

        Font currentFont = this.getFont();
        Font newFont = new Font(currentFont.getFontName(), currentFont.getStyle(),
                                9);
        this.setFont(newFont);

        /**
         * find the maximum length of all columns' names
         */
        int maxColumnNameLength = ((String) columnNames[0]).length();
        int length;
        int i = 0;
        for (Object columnName : columnNames) {
            length = ((String) columnName).length();
            if (length > maxColumnNameLength) {
                maxColumnNameLength = length;
            }
            i++;
        }

        /**
         * set table's header preferences
         */
        JTableHeader header = this.getTableHeader();
        header.setReorderingAllowed(false);
        header.setPreferredSize(new Dimension(size, maxColumnNameLength * 6));
        header.setFont(newFont);

        /**
         * set table header columns' preferences
         */
        Enumeration<TableColumn>
                tableColumns = header.getColumnModel().getColumns();
        if (tableColumns.hasMoreElements()) {
            tableColumns.nextElement().setResizable(true);
        } while (tableColumns.hasMoreElements()) {
            tableColumns.nextElement().setResizable(false);
        }

        /**
         * set columns' preferred width
         */
        tableColumns = this.getColumnModel().getColumns();
        if (tableColumns.hasMoreElements()) {
            tableColumns.nextElement().setPreferredWidth(70);
        } while (tableColumns.hasMoreElements()) {
            tableColumns.nextElement().setPreferredWidth(size);
        }

        /**
         * calculate mean and divider values if it was asked for
         */
        if (meanDivider) {
            calcMeanDividerValues(rowData);
        }
    }

    /**
     * Overrides <code>TableValues</code>'s <code>prepareRenderer</code> method.
     * Prepares the renderer by querying the data model for the value and
     * selection state of the cell at <code>row</code>, <code>column</code>.
     * Returns the component (may be a <code>Component</code>
     * or a <code>JComponent</code>) under the event location.
     *
     * @param renderer  the <code>TableCellRenderer</code> to prepare
     * @param row       the row of the cell to render, where 0 is the first row
     * @param column    the column of the cell to render,
     *			where 0 is the first column
     * @return          the <code>Component</code> under the event location
     *
     * @see TableValues#prepareRenderer(TableCellRenderer renderer, int row, int column)
     */
    public Component prepareRenderer(TableCellRenderer renderer, int row,
                                     int column) {
        Component component = super.prepareRenderer(renderer, row, column);
        if (column != 0) {
            /**
             * this cell is not on the first column, i.e., contains a gene
             * expression value (not a gene name)
             */
            float value = ((Float)this.getValueAt(row, column)).floatValue();
            if (value == meanValue) {
                /**
                 * this cell's value equals the mean value, so it's painted black
                 */
                component.setBackground(HeatmapColorScheme.getMeanValueColor());
                component.setForeground(HeatmapColorScheme.getMeanValueColor());
            } else {
                /**
                 * this cell's value is different from the mean value; it'll
                 * be painted green or red; find the green or red color
                 * component <code>rgb</code>
                 */
                int rgb = Math.round(Math.abs(value - meanValue) / divider);

                /**
                 * limit the red/green color component, so that it is not
                 * higher than 255 nor lower than 0 (since rgb components'
                 * values always range between 0 and 255)
                 */
                if (rgb > 255) {
                    rgb = 255;
                } else
                if (rgb < 0) {
                    rgb = 0;
                }

                Color color = null;
                if (value > meanValue) {
                    /**
                     * the actual cell's value is higher than the mean value;
                     * so, the cell is colored red; red component value was
                     * calculated before:
                     * Red = <code>rgb</code>, Green = 0, Blue = 0
                     */
                    color = HeatmapColorScheme.getColorAboveMean(rgb);
                } else {
                    /**
                     * the actual cell's value is lower than the mean value;
                     * so, the cell is colored green; green component value was
                     * calculated before:
                     * Red = 0, Green = <code>rgb</code>, Blue = 0
                     */
                    color = HeatmapColorScheme.getColorBelowMean(rgb);
                }

                /**
                 * turn the prepared color brighter and set background and
                 * foreground colors with the prepared color
                 */
                for (int i = 0; i < 4; i++) {
                    color = color.brighter();
                }
                component.setBackground(color);
                component.setForeground(color);
            }
        } else
        /**
         * this cell is on the first column (it's one of those which
         * contain genes' names)
         */
        if (!isCellSelected(row, column)) {
            /**
             * this cell is on the first column and it is not selected by
             * the user; so, the background is set to default
             */
            component.setBackground(getBackground());
        }

        return component;
    }

    /**
     * Calculates mean and divider values for the given row data. These values
     * are used by <code>prepareRenderer</code> method to calculate the green
     * or red color component to paint the cell.
     * <p>
     * Mean value is calculated like this:<br/>
     * <b>meanValue = (minValue + maxValue) / 2</b><br/>
     * where minValue is the minimum value of all the expression matrix's values
     * and maxValue is the maximum value of all.
     * </p>
     * <p>
     * Divider value is calculated like this:<br/>
     * <b>divider = (meanValue-minValue) / 255</b><br/>
     * where minValue, maxValue and meanValue are the same mentioned above.
     * </p>
     *
     * @param rowData the <code>Object[][]</code> row data that includes all
     *                gene expression values, from which the mean and divider
     *                values are calculated
     */
    private void calcMeanDividerValues(Object[][] rowData) {
        float minValue = ((Float) rowData[0][1]).floatValue();
        float maxValue = minValue;
        float value;

        /**
         * find minimum and maximum values of the row data values
         */
        for (Object[] row : rowData) {
            for (int i = 1; i < row.length; i++) {
                value = ((Float) row[i]).floatValue();
                if (value < minValue) {
                    minValue = value;
                }
                if (value > maxValue) {
                    maxValue = value;
                }
            }
        }

        /**
         * calculate mean and divider values
         */
        meanValue = (minValue + maxValue) / 2;
        divider = (meanValue - minValue) / 255;
    }

    /**
     * Exports this table to an image file. Allows the user to select
     * the image's filepath in a file chooser dialog and choose the image
     * format (PNG or JPG).
     */
    public void exportToImage() {
        if (parent == null) {
            return;
        }
        JFileChooser pathFileChooser = new JFileChooser();
        try {
            /**
             * initialize the file chooser with the intended file filters
             * (one for png files, and another for svg files)
             */
            if (parent instanceof BiggestsFrame) {
                pathFileChooser.addChoosableFileFilter(new FileFilterDirectory(
                        ((BiggestsFrame) parent).getPathsToFolders(), ".png"));
                pathFileChooser.addChoosableFileFilter(new FileFilterDirectory(
                        ((BiggestsFrame) parent).getPathsToFolders(), ".jpg"));
            } else {
                pathFileChooser.addChoosableFileFilter(new FileFilterExtension(
                        ".png"));
                pathFileChooser.addChoosableFileFilter(new FileFilterExtension(
                        ".jpg"));
            }
        } catch (Exception ex) {
            ex.printStackTrace();
        }

        /**
         * turn off accept all file filter
         */
        pathFileChooser.setAcceptAllFileFilterUsed(false);

        /**
         * show the save dialog to the user
         */
        int option = pathFileChooser.showSaveDialog(parent);

        if (option == JFileChooser.APPROVE_OPTION) {
            /**
             * user hit approve option; get the specified path for the file
             */
            String newPath = pathFileChooser.getSelectedFile().getPath();
            String extension = "";
            if (parent instanceof BiggestsFrame) {
                extension = ((FileFilterDirectory) pathFileChooser.
                             getFileFilter()).getValidFileExtension();
                newPath = ((FileFilterDirectory) pathFileChooser.
                           getFileFilter()).checkAndAddFileExtension(newPath);
            } else {
                extension = ((FileFilterExtension) pathFileChooser.
                             getFileFilter()).getValidFileExtension();
                newPath = ((FileFilterExtension) pathFileChooser.
                           getFileFilter()).checkAndAddFileExtension(newPath);
            }
            File newFile = new File(newPath);
            if (newFile.exists()) {
                if (newFile.isDirectory()) {
                    JOptionPane.showMessageDialog(parent,
                                                  "Unable to save image." +
                                                  "\nYou didn't specify a filename.",
                                                  "Filename missing",
                                                  JOptionPane.ERROR_MESSAGE);
                    return;
                }
                if (newFile.isFile()) {
                    int opt = JOptionPane.showOptionDialog(parent,
                            "The specified file already exists." +
                            "\nOverwrite?",
                            "Overwrite?",
                            JOptionPane.YES_NO_OPTION,
                            JOptionPane.QUESTION_MESSAGE,
                            null, null, JOptionPane.YES_OPTION);
                    if (opt == JOptionPane.NO_OPTION) {
                        /**
                         * the user does not want to overwrite the file
                         */
                        return;
                    }
                }
            }

            try {
                int locationX = this.getTableHeader().getLocation().x;
                int locationY = this.getTableHeader().getLocation().y;
                int borderWidth = 0;
                Color oldColor = this.getBackground();
                this.setBackground(Color.white);
                this.getTableHeader().setBackground(Color.white);
                this.getParent().setBackground(Color.white);
                if (this.getParent() instanceof JComponent) {
                    Border border = ((JComponent)this.getParent()).getBorder();
                    if (border != null) {
                        borderWidth = border.getBorderInsets(this.getParent()).
                                      left +
                                      border.getBorderInsets(this.getParent()).
                                      right;
                    }
                }
                try {
                    BiggestsIO.exportComponentToImage(this.getParent(),
                            locationX, locationY,
                            this.getWidth() + borderWidth,
                            this.getParent().getHeight() - locationY,
                            extension, newPath);
                } catch (Exception ex) {
                    BiggestsIO.exportComponentToImage(this.getParent(),
                            extension, newPath);
                }
                this.setBackground(oldColor);
                this.getTableHeader().setBackground(oldColor);
                this.getParent().setBackground(oldColor);
            } catch (Exception ex) {
                ex.printStackTrace();
                JOptionPane.showMessageDialog(parent, "Unable to export " +
                                              "expression matrix as image.",
                                              "Export as image failed",
                                              JOptionPane.ERROR_MESSAGE);
            }
        }
    }

    /**
     * Returns the mean value of the float values in the table.
     *
     * @return the <code>float</code> mean of the float values in the table
     */
    public float getMeanValue() {
        return meanValue;
    }

    /**
     * Returns the divider value for the float values in the table, that is,
     * the value that will be used to divide the difference between the
     * true float value of a cell and the <code>meanValue</code> to find the red or green
     * color component to colorize the cell.
     *
     * @return the <code>float</code> divider value
     */
    public float getDivider() {
        return divider;
    }

    /**
     * Handles a mouse click in this table of colors. If the click is
     * in the mouse's left button, calls <code>TableValues</code>
     * <code>this_mouseClicked</code> method to display the terms
     * annotated with the gene in the clicked row of the table.
     *
     * If the click is in the mouse's right button, shows 'Export to image'
     * popup menu.
     *
     * @param e the <code>MouseEvent</code> fired when the user clicks this table
     */
    public void this_mouseClicked(MouseEvent e) {
        if (e.getButton() == MouseEvent.BUTTON1) {
            super.this_mouseClicked(e);
        } else {
            JPopupMenu popup = new JPopupMenu();
            JMenuItem item = new JMenuItem("Export as image");
            item.addActionListener(new ItemHandler());
            popup.add(item);
            popup.show(this, e.getPoint().getLocation().x,
                       e.getPoint().getLocation().y);
        }
    }

    /**
     * <p>Description: Implements an ActionListener for 'Export as image'
     *                 context menu item.
     * </p>
     */
    private class ItemHandler implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            exportToImage();
        }
    }
}
