package biggests.tables;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import biggests.utils.Arrow;

/**
 * <p>Title: Table Sorter - a model for sorting tables</p>
 *
 * <p>Description:  TableSorter is a decorator for TableModels adding
 *                  sorting functionality to a supplied TableModel.
 *                  TableSorter does not store or copy the data in its
 *                  TableModel; instead it maintains a map from the row
 *                  indexes of the view to the row indexes of the model.
 *                  As requests are made of the sorter (like
 *                  getValueAt(row, col)) they are passed to the underlying
 *                  model after the row numbers have been translated via
 *                  the internal mapping array. This way, the TableSorter
 *                  appears to hold another copy of the table with the rows
 *                  in a different order.<br/>
 *
 *                  TableSorter registers itself as a listener to the
 *                  underlying model, just as the JTable itself would.
 *                  Events received from the model are examined, sometimes
 *                  manipulated (typically widened), and then passed on to
 *                  the TableSorter's listeners (typically the JTable).
 *                  If a change to the model has invalidated the order of
 *                  TableSorter's rows, a note of this is made and the
 *                  sorter will resort the rows the next time a value is
 *                  requested.<br/>
 *
 *                  When the tableHeader property is set, either by using
 *                  the setTableHeader() method or the two argument
 *                  constructor, the table header may be used as a complete
 *                  UI for TableSorter. The default renderer of the
 *                  tableHeader is decorated with a renderer that indicates
 *                  the sorting status of each column. In addition, a mouse
 *                  listener is installed with the following behavior:
 * <ul>
 * <li>             Mouse-click: Clears the sorting status of all other
 *                  columns and advances the sorting status of that column
 *                  through three values: {NOT_SORTED, ASCENDING, DESCENDING}
 *                  (then back to NOT_SORTED again).
 * <li>             SHIFT-mouse-click: Clears the sorting status of all
 *                  other columns and cycles the sorting status of the
 *                  column through the same three values, in the opposite
 *                  order: {NOT_SORTED, DESCENDING, ASCENDING}.
 * <li>             CONTROL-mouse-click and CONTROL-SHIFT-mouse-click: as
 *                  above except that the changes to the column do not
 *                  cancel the statuses of columns that are already sorting
 *                  - giving a way to initiate a compound sort.
 * </ul>
 *                  This is a long overdue rewrite of a class of the same
 *                  name that first appeared in the swing table demos in 1997.
 * </p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * Version of 02/27/04
 *
 * @author Philip Milne
 * @author Brendon McLean
 * @author Dan van Enckevort
 * @author Parwinder Sekhon
 * @version 2.0
 */
public class TableSorter extends AbstractTableModel {
    /**
     * The table model.
     */
    protected TableModel tableModel;

    /**
     * The status that indicates that a specific column is sorted
     * in descending order.
     */
    public static final int DESCENDING = -1;

    /**
     * The status that indicates that a specific column is not sorted.
     */
    public static final int NOT_SORTED = 0;

    /**
     * The status that indicates that a specific column is sorted
     * in ascending order.
     */
    public static final int ASCENDING = 1;

    /**
     * Sorting directive.
     */
    private static Directive EMPTY_DIRECTIVE = new Directive( -1, NOT_SORTED);

    /**
     * The comparator for specific comparable objects.
     */
    public static final Comparator COMPARABLE_COMPARATOR = new Comparator() {
        public int compare(Object o1, Object o2) {
            return ((Comparable) o1).compareTo(o2);
        }
    };

    /**
     * The lexical comparator.
     */
    public static final Comparator LEXICAL_COMPARATOR = new Comparator() {
        public int compare(Object o1, Object o2) {
            return o1.toString().compareTo(o2.toString());
        }
    };

    /**
     * View to model indexes map.
     */
    private Row[] viewToModel;

    /**
     * Model to view indexes map.
     */
    private int[] modelToView;

    /**
     * The header of the table.
     * Set to protected access to be used by subclasses (by Joana Gonalves
     * on 24-11-2006)
     */
    protected JTableHeader tableHeader;

    /**
     * The mouse listener of the table.
     * Set to protected access to be used by subclasses (by Joana Gonalves
     * on 24-11-2006).
     */
    protected MouseListener mouseListener;

    /**
     * The table model listener.
     */
    private TableModelListener tableModelListener;

    /**
     * Hash map for the columns' comparators.
     * Set to protected access to be used by subclasses (by Joana Gonalves
     * on 24-11-2006).
     */
    protected HashMap<Class, Comparator> columnComparators =
            new HashMap<Class, Comparator>();

    /**
     * The directives to sort columns.
     */
    private ArrayList<Directive> sortingColumns =
            new ArrayList<Directive>();

    /**
     * Creates a new <code>TableSorter</code> with default mouse and
     * table model listeners.
     */
    public TableSorter() {
        this.mouseListener = new MouseHandler();
        this.tableModelListener = new TableModelHandler();
    }

    /**
     * Creates a new <code>TableSorter</code> with the given table model.
     *
     * @param tableModel <code>TableModel</code> the model for this table
     */
    public TableSorter(TableModel tableModel) {
        this();
        setTableModel(tableModel);
    }

    /**
     * Creates a new <code>TableSorter</code> with the given table
     * model and header.
     *
     * @param tableModel the <code>TableModel</code>
     * @param tableHeader the <code>JTableHeader</code>
     */
    public TableSorter(TableModel tableModel, JTableHeader tableHeader) {
        this();
        setTableHeader(tableHeader);
        setTableModel(tableModel);
    }

    /**
     * Clears sorting state.
     */
    private void clearSortingState() {
        viewToModel = null;
        modelToView = null;
    }

    /**
     * Returns the table model.
     *
     * @return TableModel
     */
    public TableModel getTableModel() {
        return tableModel;
    }

    /**
     * Sets the table model.
     *
     * @param tableModel the <code>TableModel</code>
     */
    public void setTableModel(TableModel tableModel) {
        if (this.tableModel != null) {
            this.tableModel.removeTableModelListener(tableModelListener);
        }

        this.tableModel = tableModel;
        if (this.tableModel != null) {
            this.tableModel.addTableModelListener(tableModelListener);
        }

        clearSortingState();
        fireTableStructureChanged();
    }

    /**
     * Returns the table header.
     *
     * @return JTableHeader the header of the table
     */
    public JTableHeader getTableHeader() {
        return tableHeader;
    }

    /**
     * Sets the table header.
     *
     * @param tableHeader the new <code>JTableHeader</code>
     */
    public void setTableHeader(JTableHeader tableHeader) {
        if (this.tableHeader != null) {
            this.tableHeader.removeMouseListener(mouseListener);
            TableCellRenderer defaultRenderer = this.tableHeader.
                                                getDefaultRenderer();
            if (defaultRenderer instanceof SortableHeaderRenderer) {
                this.tableHeader.setDefaultRenderer(((SortableHeaderRenderer)
                        defaultRenderer).tableCellRenderer);
            }
        }
        this.tableHeader = tableHeader;
        if (this.tableHeader != null) {
            this.tableHeader.addMouseListener(mouseListener);
            this.tableHeader.setDefaultRenderer(
                    new SortableHeaderRenderer(this.tableHeader.
                                               getDefaultRenderer()));
        }
    }

    /**
     * Checks if the table is sorting.
     * Returns <code>true</code> if table is sorting,
     * <code>false</code> otherwise.
     *
     * @return boolean <code>true</code> if table is sorting,
     * <code>false</code> otherwise
     */
    public boolean isSorting() {
        return sortingColumns.size() != 0;
    }

    /**
     * Returns the sorting directive for the given <code>column</code>.
     *
     * @param column <code>int</code> the index of the column
     * @return Directive the sorting directive of the given column
     */
    private Directive getDirective(int column) {
        for (int i = 0; i < sortingColumns.size(); i++) {
            Directive directive = sortingColumns.get(i);
            if (directive.column == column) {
                return directive;
            }
        }
        return EMPTY_DIRECTIVE;
    }

    /**
     * Returns the sorting status of the given <code>column</code>:<br/>
     * <ul>
     * <li>DESCENDING
     * <li>NOT_SORTED
     * <li>ASCENDING
     * </ul>
     *
     * @param column <code>int</code> the index of the column
     * @return int the sorting status of the given column
     */
    public int getSortingStatus(int column) {
        return getDirective(column).direction;
    }

    /**
     * Called when sorting status changes. Clears sorting state, fires
     * the event table data changed and repaint the table header.
     */
    private void sortingStatusChanged() {
        clearSortingState();
        fireTableDataChanged();
        if (tableHeader != null) {
            tableHeader.repaint();
        }
    }

    /**
     * Sets the sorting status for a given <code>column</code>.
     *
     * @param column <code>int</code> the column to set sorting status
     * @param status <code>int</code> the new sorting status for the given column
     */
    public void setSortingStatus(int column, int status) {
        Directive directive = getDirective(column);
        if (directive != EMPTY_DIRECTIVE) {
            sortingColumns.remove(directive);
        }
        if (status != NOT_SORTED) {
            sortingColumns.add(new Directive(column, status));
        }
        sortingStatusChanged();
    }

    /**
     * Returns the Icon corresponding to the sorting status of the
     * given <code>column</code>. The icon will have the specified
     * <code>size</code>.
     *
     * @param column <code>int</code> the column to return the sorting
     *        status icon for
     * @param size <code>int</code> the size of the sorting status icon
     * @return Icon
     */
    protected Icon getHeaderRendererIcon(int column, int size) {
        Directive directive = getDirective(column);
        if (directive == EMPTY_DIRECTIVE) {
            return null;
        }
        return new Arrow(directive.direction == DESCENDING, size,
                         sortingColumns.indexOf(directive));
    }

    /**
     * Cancels sorting of all columns and fires sorting status changed event.
     */
    private void cancelSorting() {
        sortingColumns.clear();
        sortingStatusChanged();
    }

    /**
     * Sets the given <code>comparator</code> for the given objects'
     * <code>type</code>.
     *
     * @param type the <code>Class</code> of objects to use the given
     *        comparator with
     * @param comparator the appropriate <code>Comparator</code> for
     *        the objects of the specified <code>type</code>
     */
    public void setColumnComparator(Class type, Comparator comparator) {
        if (comparator == null) {
            columnComparators.remove(type);
        } else {
            columnComparators.put(type, comparator);
        }
    }

    /**
     * Returns the appropriate comparator for the objects in the
     * given <code>column</code>.
     *
     * @param column <code>int</code> the column for which the appropriate
     *        comparator should be returned
     * @return Comparator the appropriate comparator for the given
     *         <code>column</code>
     */
    protected Comparator getComparator(int column) {
        Class columnType = tableModel.getColumnClass(column);
        Comparator comparator = (Comparator) columnComparators.get(columnType);

        if (comparator != null) {
            return comparator;
        }
        if (Comparable.class.isAssignableFrom(columnType)) {
            return COMPARABLE_COMPARATOR;
        }
        return LEXICAL_COMPARATOR;
    }

    /**
     * Returns the view to model mapping.
     * getViewToModel()[viewIndex].modelIndex gives the modelIndex for
     * a known viewIndex.
     *
     * @return Row[] the view to model indexes mapping
     */
    private Row[] getViewToModel() {
        if (viewToModel == null) {
            int tableModelRowCount = tableModel.getRowCount();
            viewToModel = new Row[tableModelRowCount];
            for (int row = 0; row < tableModelRowCount; row++) {
                viewToModel[row] = new Row(row);
            }

            if (isSorting()) {
                Arrays.sort(viewToModel);
            }
        }
        return viewToModel;
    }

    /**
     * Returns the model index for a given <code>viewIndex</code>.
     *
     * @param viewIndex <code>int</code> the index that is shown in
     *        the table
     * @return int the model index: the index of the row in the table
     *        model, which may be different from the view index, if some
     *        column is being sorted
     */
    public int modelIndex(int viewIndex) {
        return getViewToModel()[viewIndex].modelIndex;
    }

    /**
     * Returns the model to view indexes mapping.
     * getModelToView[modelIndex] gives the view index for the given modelIndex.
     *
     * @return int[] the model to view indexes mapping
     */
    private int[] getModelToView() {
        if (modelToView == null) {
            int n = getViewToModel().length;
            modelToView = new int[n];
            for (int i = 0; i < n; i++) {
                modelToView[modelIndex(i)] = i;
            }
        }
        return modelToView;
    }

    /**
     * Returns the number of rows in the table model.
     *
     * @return int the number of rows in the table model
     */
    public int getRowCount() {
        return (tableModel == null) ? 0 : tableModel.getRowCount();
    }

    /**
     * Returns the number of columns in the table model.
     *
     * @return int the number of columns in the table model
     */
    public int getColumnCount() {
        return (tableModel == null) ? 0 : tableModel.getColumnCount();
    }

    /**
     * Returns the name of the given <code>column</code>.
     *
     * @param column <code>int</code> the index of the intended column
     * @return String the name of the given column
     */
    public String getColumnName(int column) {
        return tableModel.getColumnName(column);
    }

    /**
     * Returns the class of the column with index <code>column</code>.<br/>
     * Changed return class from <code>Class</code> to <code>Class<?></code> so
     * this method overrides <code>AbstractTableModel</code>'s
     * <code>getColumnClass</code> method (by Joana Gonalves).
     *
     * @param column <code>int</code> value: the column being queried
     * @return the <code>Class</code> of the column at index <code>column</code>
     */
    public Class<?> getColumnClass(int column) {
        return tableModel.getColumnClass(column);
    }

    /**
     * Checks if the cell with the given <code>row</code> and
     * <code>column</code> indexes is editable.
     *
     * @param row <code>int</code> the row index of the cell
     * @param column <code>int</code> the column index of the cell
     * @return boolean <code>true</code> if the cell at (<code>row</code>,
     *         <code>column</code>) is editable, <code>false</code> otherwise
     */
    public boolean isCellEditable(int row, int column) {
        return tableModel.isCellEditable(modelIndex(row), column);
    }

    /**
     * Returns the element in the cell at <code>row</code> and
     * <code>column</code> indexes.
     *
     * @param row <code>int</code> the row index of the cell
     * @param column <code>int</code> the column index of the cell
     * @return Object the element in the cell at (<code>row</code>,
     *         <code>column</code>)
     */
    public Object getValueAt(int row, int column) {
        return tableModel.getValueAt(modelIndex(row), column);
    }

    /**
     * Sets the element of the cell with <code>row</code> and
     * <code>column</code>) indexes.
     *
     * @param aValue <code>Object</code> the new element for the cell
     * @param row <code>int</code> the row index of the cell
     * @param column <code>int</code> the column index of the cell
     */
    public void setValueAt(Object aValue, int row, int column) {
        tableModel.setValueAt(aValue, modelIndex(row), column);
    }

    /**
     * <p>Description: Implements an object that holds the index of a
     *                 specific row in a certain table model.</p>
     */
    private class Row implements Comparable {
        /**
         * The index of the row in a specific table model.
         */
        private int modelIndex;

        /**
         * Creates a new Row with the given model index.
         *
         * @param index <code>int</code> the index of the row in the
         *        specific table model
         */
        public Row(int index) {
            this.modelIndex = index;
        }

        /**
         * Compares this row to <code>o</code>.
         * Returns a negative integer, 0 or a positive integer if this row
         * is less than, equal to or greater than <code>o</code>.
         *
         * @param o the <code>Object</code> to be compared with this
         *        <code>Row</code>
         * @return int the result of the comparation
         */
        public int compareTo(Object o) {
            int row1 = modelIndex;
            int row2 = ((Row) o).modelIndex;

            for (Iterator it = sortingColumns.iterator(); it.hasNext(); ) {
                Directive directive = (Directive) it.next();
                int column = directive.column;
                Object o1 = tableModel.getValueAt(row1, column);
                Object o2 = tableModel.getValueAt(row2, column);

                int comparison = 0;
                // Define null less than everything, except null.
                if (o1 == null && o2 == null) {
                    comparison = 0;
                } else if (o1 == null) {
                    comparison = -1;
                } else if (o2 == null) {
                    comparison = 1;
                } else {
                    comparison = getComparator(column).compare(o1, o2);

                }
                if (comparison != 0) {
                    return directive.direction == DESCENDING ? -comparison :
                            comparison;
                }
            }
            return 0;
        }
    }


    /**
     * <p>Description: The table model handler.</p>
     */
    private class TableModelHandler implements TableModelListener {
        /**
         * Handles the tabled changed event and passes the event along.
         *
         * @param e <code>TableModelEvent</code> fired when a change
         *        in the table model happens
         */
        public void tableChanged(TableModelEvent e) {
            // If we're not sorting by anything, just pass the event along.
            if (!isSorting()) {
                clearSortingState();
                fireTableChanged(e);
                return;
            }

            // If the table structure has changed, cancel the sorting; the
            // sorting columns may have been either moved or deleted from
            // the model.
            if (e.getFirstRow() == TableModelEvent.HEADER_ROW) {
                cancelSorting();
                fireTableChanged(e);
                return;
            }

            // We can map a cell event through to the view without widening
            // when the following conditions apply:
            //
            // a) all the changes are on one row (e.getFirstRow() == e.getLastRow()) and,
            // b) all the changes are in one column (column != TableModelEvent.ALL_COLUMNS) and,
            // c) we are not sorting on that column (getSortingStatus(column) == NOT_SORTED) and,
            // d) a reverse lookup will not trigger a sort (modelToView != null)
            //
            // Note: INSERT and DELETE events fail this test as they have column == ALL_COLUMNS.
            //
            // The last check, for (modelToView != null) is to see if modelToView
            // is already allocated. If we don't do this check; sorting can become
            // a performance bottleneck for applications where cells
            // change rapidly in different parts of the table. If cells
            // change alternately in the sorting column and then outside of
            // it this class can end up re-sorting on alternate cell updates -
            // which can be a performance problem for large tables. The last
            // clause avoids this problem.
            int column = e.getColumn();
            if (e.getFirstRow() == e.getLastRow()
                && column != TableModelEvent.ALL_COLUMNS
                && getSortingStatus(column) == NOT_SORTED
                && modelToView != null) {
                int viewIndex = getModelToView()[e.getFirstRow()];
                fireTableChanged(new TableModelEvent(TableSorter.this,
                        viewIndex, viewIndex,
                        column, e.getType()));
                return;
            }

            // Something has happened to the data that may have invalidated the row order.
            clearSortingState();
            fireTableDataChanged();
            return;
        }
    }


    /**
     * <p>Description: Mouse Handler.</p>
     */
    private class MouseHandler extends MouseAdapter {
        /**
         * Handles the event fired when the mouse is clicked on the table.
         * Checks in which column was the mouse clicked, cancels sorting
         * in all columns and performs sorting only in the specified one.
         *
         * @param e <code>MouseEvent</code> fired when the mouse is clicked
         *        on the table
         */
        public void mouseClicked(MouseEvent e) {
            JTableHeader h = (JTableHeader) e.getSource();
            TableColumnModel columnModel = h.getColumnModel();
            int viewColumn = columnModel.getColumnIndexAtX(e.getX());
            /**
             * added by Joana P. Gonalves
             */
            if (viewColumn == -1) {
                return;
            }

            int column = columnModel.getColumn(viewColumn).getModelIndex();
            if (column != -1) {
                int status = getSortingStatus(column);
                if (!e.isControlDown()) {
                    cancelSorting();
                }
                // Cycle the sorting states through {NOT_SORTED, ASCENDING, DESCENDING} or
                // {NOT_SORTED, DESCENDING, ASCENDING} depending on whether shift is pressed.
                status = status + (e.isShiftDown() ? -1 : 1);
                status = (status + 4) % 3 - 1; // signed mod, returning {-1, 0, 1}
                setSortingStatus(column, status);
            }
        }
    }


    /**
     * <p>Description: Implements a renderer for this table's cells.</p>
     */
    private class SortableHeaderRenderer implements TableCellRenderer {
        /**
         * The table cells' renderer.
         */
        private TableCellRenderer tableCellRenderer;

        /**
         * Creates a new SortableHeaderRenderer with the given
         * <code>tableCellRenderer</code>.
         *
         * @param tableCellRenderer <code>TableCellRenderer</code>
         */
        public SortableHeaderRenderer(TableCellRenderer tableCellRenderer) {
            this.tableCellRenderer = tableCellRenderer;
        }

        /**
         * Returns the renderer component for the cell at <code>row</code>
         * and <code>column</code> indexes.
         *
         * @param table <code>JTable</code the table
         * @param value <code>Object</code>
         * @param isSelected <code>boolean</code> indicates if cell is
         *        selected or not
         * @param hasFocus <code>boolean</code> indicates if cell has focus
         * @param row <code>int</code> the index of the cell's row
         * @param column <code>int</code> the index of the cell's column
         * @return Component the renderer component for the specified cell
         */
        public Component getTableCellRendererComponent(JTable table,
                Object value,
                boolean isSelected,
                boolean hasFocus,
                int row,
                int column) {
            Component c = tableCellRenderer.getTableCellRendererComponent(table,
                    value, isSelected, hasFocus, row, column);
            if (c instanceof JLabel) {
                JLabel l = (JLabel) c;
                l.setHorizontalTextPosition(JLabel.LEFT);
                int modelColumn = table.convertColumnIndexToModel(column);
                l.setIcon(getHeaderRendererIcon(modelColumn,
                                                l.getFont().getSize()));
            }
            return c;
        }
    }


    /**
     * <p>Description: A sorting directive.</p>
     */
    private static class Directive {
        /**
         * The column index.
         */
        private int column;

        /**
         * The sorting direction.
         */
        private int direction;

        /**
         * Creates a new sorting directive, for the given <code>column</code>
         * with the given <code>direction</code>.
         *
         * @param column <code>int</code> the index of the column
         * @param direction <code>int</code> the sorting direction
         */
        public Directive(int column, int direction) {
            this.column = column;
            this.direction = direction;
        }
    }
}
